import decimal
import json as _json
import os
import sys
from pathlib import Path
from typing import Any, Callable, Optional

import logging
import numpy as np
import pandas as pd
import yaml
import requests
from dotenv import dotenv_values  # type:ignore
from huggingface_hub import hf_hub_download as _hf_hub_download
from huggingface_hub import snapshot_download as _snapshot_download
from .dao import Dao
from .config import FlojoyConfig, logger

from .node_init import NodeInit, NodeInitService
import keyring
import base64


__all__ = [
    "send_to_socket",
    "get_env_var",
    "set_env_var",
    "delete_env_var",
    "get_credentials",
    "hf_hub_download",
    "snapshot_download",
    "send_to_socket",
    "hf_hub_download",
    "snapshot_download",
    "clear_flojoy_memory",
]

FLOJOY_DIR = ".flojoy"


if sys.platform == "win32":
    FLOJOY_CACHE_DIR = os.path.realpath(os.path.join(os.environ["APPDATA"], FLOJOY_DIR))
else:
    FLOJOY_CACHE_DIR = os.path.realpath(os.path.join(os.environ["HOME"], FLOJOY_DIR))


# Make as a function to mock at test-time
def _get_hf_hub_cache_path() -> str:
    return os.path.join(FLOJOY_CACHE_DIR, "cache", "hf_hub")


def hf_hub_download(*args, **kwargs):
    if "cache_dir" not in kwargs:
        kwargs["cache_dir"] = _get_hf_hub_cache_path()
    else:
        if kwargs["cache_dir"] != _get_hf_hub_cache_path():
            raise ValueError(
                f"Attempted to override cache_dir parameter, received {kwargs['cache_dir']} while the only alloed value is {_get_hf_hub_cache_path()}"
            )
    return _hf_hub_download(*args, **kwargs)


def snapshot_download(*args, **kwargs):
    if "cache_dir" not in kwargs:
        kwargs["cache_dir"] = _get_hf_hub_cache_path()
    else:
        if kwargs["cache_dir"] != _get_hf_hub_cache_path():
            raise ValueError(
                f"Attempted to override cache_dir parameter, received {kwargs['cache_dir']} while the only alloed value is {_get_hf_hub_cache_path()}"
            )
    return _snapshot_download(*args, **kwargs)


env_vars = dotenv_values("../.env")
port = env_vars.get("VITE_BACKEND_PORT", "5392")
BACKEND_URL = os.environ.get("BACKEND_URL", f"http://127.0.0.1:{port}")


def set_offline():
    """
    Sets the is_offline flag to True, which means that results will not be sent to the backend via HTTP.
    Mainly used for precompilation
    """
    FlojoyConfig.get_instance().is_offline = True


def set_online():
    """
    Sets the is_offline flag to False, which means that results will be sent to the backend via HTTP.
    """
    FlojoyConfig.get_instance().is_offline = False


def set_debug_on():
    """
    Sets the print_on flag to True, which means that the print statements will be executed.
    """
    logger.setLevel(logging.DEBUG)


def set_debug_off():
    """
    Sets the print_on flag to False, which means that the print statements will not be executed.
    """
    logger.setLevel(logging.INFO)


def clear_flojoy_memory():
    Dao.get_instance().clear_job_results()
    Dao.get_instance().clear_small_memory()
    Dao.get_instance().clear_node_init_containers()


def send_to_socket(data: str):
    if FlojoyConfig.get_instance().is_offline:
        return
    logger.debug("posting data to socket:", f"{BACKEND_URL}/worker_response")
    requests.post(f"{BACKEND_URL}/worker_response", json=data)


class PlotlyJSONEncoder(_json.JSONEncoder):
    """
    Meant to be passed as the `cls` kwarg to json.dumps(obj, cls=..)
    See PlotlyJSONEncoder.default for more implementation information.
    Additionally, this encoder overrides nan functionality so that 'Inf',
    'NaN' and '-Inf' encode to 'null'. Which is stricter JSON than the Python
    version.
    """

    def coerce_to_strict(self, const: Any):
        """
        This is used to ultimately *encode* into strict JSON, see `encode`
        """
        # before python 2.7, 'true', 'false', 'null', were include here.
        if const in ("Infinity", "-Infinity", "NaN"):
            return None
        else:
            return const

    def encode(self, o: Any):
        """
        Load and then dump the result using parse_constant kwarg
        Note that setting invalid separators will cause a failure at this step.
        """
        # this will raise errors in a normal-expected way
        encoded_o = super(PlotlyJSONEncoder, self).encode(o)
        # Brute force guessing whether NaN or Infinity values are in the string
        # We catch false positive cases (e.g. strings such as titles, labels etc.)
        # but this is ok since the intention is to skip the decoding / reencoding
        # step when it's completely safe

        if not ("NaN" in encoded_o or "Infinity" in encoded_o):
            return encoded_o

        # now:
        #    1. `loads` to switch Infinity, -Infinity, NaN to None
        #    2. `dumps` again so you get 'null' instead of extended JSON
        try:
            new_o = _json.loads(encoded_o, parse_constant=self.coerce_to_strict)
        except ValueError:
            # invalid separators will fail here. raise a helpful exception
            raise ValueError(
                "Encoding into strict JSON failed. Did you set the separators "
                "valid JSON separators?"
            )
        else:
            return _json.dumps(
                new_o,
                sort_keys=self.sort_keys,
                indent=self.indent,
                separators=(self.item_separator, self.key_separator),
            )

    def default(self, obj: dict[str, Any]):
        """
        Accept an object (of unknown type) and try to encode with priority:
        1. builtin:     user-defined objects
        2. sage:        sage math cloud
        3. pandas:      dataframes/series
        4. numpy:       ndarrays
        5. datetime:    time/datetime objects
        Each method throws a NotEncoded exception if it fails.
        The default method will only get hit if the object is not a type that
        is naturally encoded by json:
            Normal objects:
                dict                object
                list, tuple         array
                str, unicode        string
                int, long, float    number
                True                true
                False               false
                None                null
            Extended objects:
                float('nan')        'NaN'
                float('infinity')   'Infinity'
                float('-infinity')  '-Infinity'
        Therefore, we only anticipate either unknown iterables or values here.
        """
        # TODO: The ordering if these methods is *very* important. Is this OK?
        encoding_methods = (
            self.encode_as_plotly,
            self.encode_as_numpy,
            self.encode_as_pandas,
            self.encode_as_datetime,
            self.encode_as_date,
            self.encode_as_list,  # because some values have `tolist` do last.
            self.encode_as_decimal,
            self.encode_as_base64,
        )
        for encoding_method in encoding_methods:
            try:
                return encoding_method(obj)
            except NotEncodable:
                pass
        return _json.JSONEncoder.default(self, obj)

    @staticmethod
    def encode_as_base64(value: bytes):
        """Attempt to convert to base64."""
        try:
            return base64.b64encode(value).decode()
        except AttributeError:
            raise NotEncodable

    @staticmethod
    def encode_as_plotly(obj: dict[str, Any]):
        """Attempt to use a builtin `to_plotly_json` method."""
        try:
            return obj.to_plotly_json()
        except AttributeError:
            raise NotEncodable

    @staticmethod
    def encode_as_list(obj):
        """Attempt to use `tolist` method to convert to normal Python list."""
        if hasattr(obj, "tolist"):
            return obj.tolist()
        else:
            raise NotEncodable

    @staticmethod
    def encode_as_pandas(obj):
        """Attempt to convert pandas.NaT"""
        if not pd:
            raise NotEncodable
        if obj is pd.NaT:
            return None
        elif isinstance(obj, pd.DataFrame):
            return obj.to_json()
        else:
            raise NotEncodable

    @staticmethod
    def encode_as_numpy(obj):
        """Attempt to convert numpy.ma.core.masked"""
        if not np:
            raise NotEncodable

        if obj is np.ma.masked:
            return float("nan")
        elif isinstance(obj, np.ndarray) and obj.dtype.kind == "M":
            try:
                return np.datetime_as_string(obj).tolist()
            except TypeError:
                pass

        raise NotEncodable

    @staticmethod
    def encode_as_datetime(obj):
        """Convert datetime objects to iso-format strings"""
        try:
            return obj.isoformat()
        except AttributeError:
            raise NotEncodable

    @staticmethod
    def encode_as_date(obj):
        """Attempt to convert to utc-iso time string using date methods."""
        try:
            time_string = obj.isoformat()
        except AttributeError:
            raise NotEncodable
        else:
            return time_string  # iso_to_plotly_time_string(time_string)

    @staticmethod
    def encode_as_decimal(obj):
        """Attempt to encode decimal by converting it to float"""
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        else:
            raise NotEncodable


class NotEncodable(Exception):
    pass


def dump_str(result: Any, limit: int | None = None):
    result_str = str(result)
    return (
        result_str
        if limit is None or len(result_str) <= limit
        else result_str[:limit] + "..."
    )


def get_flojoy_root_dir() -> str:
    home = str(Path.home())
    path = os.path.join(home, ".flojoy/flojoy.yaml")
    stream = open(path, "r")
    yaml_dict = yaml.load(stream, Loader=yaml.FullLoader)
    root_dir = ""

    if isinstance(yaml_dict, str):
        root_dir = yaml_dict.split(":")[1]
    else:
        root_dir = yaml_dict["PATH"]

    return root_dir


def get_env_var(key: str) -> Optional[str]:
    return keyring.get_password("flojoy", key)


def set_env_var(key: str, value: str):
    keyring.set_password("flojoy", key, value)
    home = str(Path.home())
    file_path = os.path.join(home, os.path.join(FLOJOY_DIR, "credentials.txt"))

    if not os.path.exists(file_path):
        with open(file_path, "w") as f:
            f.write(key)
        return

    with open(file_path, "r") as f:
        keys = f.read().strip().split(",")
        if key not in keys:
            keys.append(key)

    with open(file_path, "w") as f:
        f.write(",".join(keys))


def delete_env_var(key: str):
    home = str(Path.home())
    file_path = os.path.join(home, os.path.join(FLOJOY_DIR, "credentials.txt"))

    if not os.path.exists(file_path):
        return

    with open(file_path, "r") as f:
        keys = f.read().strip().split(",")

    if key not in keys:
        return

    keys.remove(key)

    with open(file_path, "w") as f:
        f.write(",".join(keys))

    keyring.delete_password("flojoy", key)


def get_credentials() -> list[dict[str, str]]:
    home = str(Path.home())
    file_path = os.path.join(home, os.path.join(FLOJOY_DIR, "credentials.txt"))

    if not os.path.exists(file_path):
        return []

    with open(file_path, "r") as f:
        keys = f.read().strip().split(",")

    credentials_list: list[dict[str, str]] = []
    for key in keys:
        val = get_env_var(key)
        if val:
            credentials_list.append({"key": key, "value": val})

    return credentials_list


def get_node_init_function(node_func: Callable) -> NodeInit:
    return NodeInitService().get_node_init_function(node_func)



import plotly.express as px
import plotly.graph_objects as go
import numpy as np
from .data_container import DataContainer
import pandas as pd
from typing import cast, Any


def data_container_to_plotly(data: DataContainer) -> dict[str, Any] | None:
    data_copy = data.copy()
    dc_type = data_copy.type
    fig = go.Figure(layout=dict(template="plotly"))
    if "x" in data_copy and isinstance(data_copy.x, dict):  # type: ignore
        data_keys = list(cast(list[str], data_copy.x.keys()))
        data_copy.x = data_copy.x[data_keys[0]]  # type: ignore

    match dc_type:
        case "Image":
            if data_copy.a is None:
                img_combined = np.stack((data_copy.r, data_copy.g, data_copy.b), axis=2)
            else:
                img_combined = np.stack(
                    (data_copy.r, data_copy.g, data_copy.b, data_copy.a), axis=2
                )
            fig = px.imshow(img=img_combined)  # type:ignore
        case "OrderedPair":
            if data_copy.x is not None and len(data_copy.x) != len(data_copy.y):
                data_copy.x = np.arange(0, len(data_copy.y), 1)
            fig = px.line(x=data_copy.x, y=data_copy.y)
        case "OrderedTriple":
            fig = px.scatter_3d(x=data_copy.x, y=data_copy.y, z=data_copy.z)
        case "Scalar":
            fig.add_trace(
                go.Indicator(
                    value=data_copy.c,
                    domain={"y": [0, 1], "x": [0, 1]},
                    number={"valueformat": "f"},
                )
            )
        case "Vector":
            df = pd.DataFrame(data_copy.v)
            fig = go.Figure(
                data=[go.Table(header=dict(values=["Vector"]), cells=dict(values=[df]))]
            )
        case "DataFrame":
            df = cast(pd.DataFrame, data_copy.m)
            fig = go.Figure(
                data=[
                    go.Table(
                        header=dict(values=list(df.columns)),
                        cells=dict(values=[df[col] for col in df.columns]),
                    )
                ]
            )
        case "Grayscale" | "Matrix":
            y_columns: np.ndarray = data_copy.m
            for i, col in enumerate(y_columns.T):
                fig.add_trace(
                    go.Scatter(
                        x=np.arange(0, col.size),
                        y=col,
                        mode="lines",
                        name=i,
                    )
                )
        case "Surface":
            fig = go.Figure(
                data=[go.Surface(x=data_copy.x, y=data_copy.y, z=data_copy.z)]
            )
        case "Plotly":
            fig = cast(go.Figure, data.fig)
        case "Bytes" | "TextBlob":
            return None
        case _:
            raise ValueError(
                f"unsupported DataContainer type passed to plotly converter function, type: '{dc_type}"
            )
    return cast(dict[str, Any], fig.to_dict())



from typing import Any
from .dao import Dao


class SmallMemory:
    """
    SmallMemory - available during jobset execution - intended to be used ONLY inside node functions
    """

    tracing_key = "ALL_MEMORY_KEYS"
    dao = Dao.get_instance()

    def clear_memory(self):
        self.dao.clear_small_memory()

    def write_to_memory(self, job_id: str, key: str, value: Any):
        memory_key = f"{job_id}-{key}"
        value_type_key = f"{memory_key}_value_type_key"
        meta_data = {}
        s = str(type(value))
        v_type = s.split("'")[1]
        match v_type:
            case "numpy.ndarray":
                array_dtype = str(value.dtype)
                meta_data["type"] = "np_array"
                meta_data["d_type"] = array_dtype
                meta_data["dimensions"] = value.shape
                self.dao.set_obj(value_type_key, meta_data)
                self.dao.set_np_array(memory_key, value)
            case "pandas.core.frame.DataFrame":
                meta_data["type"] = "pd_dframe"
                self.dao.set_obj(value_type_key, meta_data)
                self.dao.set_pandas_dataframe(memory_key, value)
            case "str" | "numpy.float64":
                meta_data["type"] = "string"
                self.dao.set_obj(value_type_key, meta_data)
                self.dao.set_str(memory_key, value)
            case "dict":
                meta_data["type"] = "dict"
                self.dao.set_obj(value_type_key, meta_data)
                self.dao.set_obj(memory_key, value)
            case _:
                raise ValueError(
                    f"SmallMemory currently does not support '{v_type}' type data!"
                )

    def read_memory(self, job_id: str, key: str):
        """
        Reads object stored in internal DB by the given key. The memory is job specific.
        """
        memory_key = f"{job_id}-{key}"
        value_type_key = f"{memory_key}_value_type_key"
        meta_data = self.dao.get_obj(value_type_key)
        if not meta_data:
            return None
        meta_type = meta_data.get("type")
        match meta_type:
            case "string":
                return self.dao.get_str(memory_key)
            case "dict":
                return self.dao.get_obj(memory_key)
            case "np_array":
                return self.dao.get_np_array(memory_key)
            case "pd_dframe":
                return self.dao.get_pd_dataframe(memory_key)
            case _:
                return None

    def delete_object(self, job_id: str, key: str):
        """
        Removes object stored in internal DB by the given key. The memory is job specific.
        """
        memory_key = f"{job_id}-{key}"
        return self.dao.delete_object(memory_key)



from typing import Callable, Any

from flojoy.parameter_types import format_param_value
from .dao import Dao


class NoInitFunctionError(Exception):
    pass


# contains value returned by a node's init function
class NodeInitContainer:
    def __init__(self, value=None):
        self.value = value

    def set(self, value):
        self.value = value

    def get(self):
        return self.value


class NodeInit:
    def __init__(self, func):
        self.func = func

    def __call__(self, node_id: str, ctrls: dict[str, Any]):
        return self.run(node_id, ctrls)

    def run(self, node_id: str, ctrls: dict[str, Any]):
        daemon_container = NodeInitService().create_init_store(node_id)

        args = {
            name: format_param_value(ctrl["value"], ctrl["type"])
            for name, ctrl in ctrls.items()
        }
        res = self.func(**args)
        if res is not None:
            daemon_container.set(res)


# Wrapper for node_init functions, maps the node to the function that will initialize it.
def node_initialization(for_node):
    def decorator(func):
        func_init = NodeInit(func)
        NodeInitService().map_node_to_init_function(for_node, func_init)
        return func_init

    return decorator


class NodeInitService:
    """
    Class that handles different tasks related to node initialization
    """

    # this method will create the storage used for the node to hold whatever it initialized.
    def create_init_store(self, node_id):
        if self.has_init_store(node_id):
            raise ValueError(f"Storage for {node_id} init object already exists!")

        Dao.get_instance().set_init_container(node_id, NodeInitContainer())
        return self.get_init_store(node_id)

    # this method will get the storage used for the node to hold whatever it initialized.
    def get_init_store(self, node_id) -> NodeInitContainer:
        store = Dao.get_instance().get_init_container(node_id)
        if store is None:
            raise ValueError(
                f"Storage for {node_id} init object has not been initialized!"
            )
        return store

    # this method will check if a node has an init store already created.
    def has_init_store(self, node_id) -> bool:
        return Dao.get_instance().has_init_container(node_id)

    # this method will map a node to a function that will initialize it.
    def map_node_to_init_function(self, node_func, node_init_func):
        if NodeInitService().has_init_store(node_func.__name__):
            raise ValueError(f"Node {node_func.__name__} already has an init store!")
        Dao.get_instance().set_init_function(node_func, node_init_func)

    # this method will get the function that will initialize a node.
    def get_node_init_function(self, node_func) -> NodeInit:
        res = Dao.get_instance().get_init_function(node_func)
        if res is None:
            raise NoInitFunctionError(
                f"Node {node_func.__name__} does not have an init function!"
            )
        return res


def get_node_init_function(node_func: Callable) -> NodeInit:
    """
    Returns the function corresponding to the init function of the specified node.
    """
    return NodeInitService().get_node_init_function(node_func)



from contextlib import contextmanager
from functools import partial, wraps
import io
from typing import Callable, Literal, Protocol
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
import logging
import multiprocessing
import multiprocessing.connection
import threading
import os


class LogPipeMode(Enum):
    """
    An enumeration of the different modes for logging pipes.

    Attributes:
        MP_SPAWN: To be used when logging from a subprocess created using the `spawn` start method of the `multiprocessing` module.
        SUBPROCESS: To be used when logging from a subprocess created using the `subprocess` module.
    """

    MP_SPAWN = auto()
    SUBPROCESS = auto()


class StreamEnum(Enum):
    """
    An enumeration of the different streams to redirect.

    Attributes:
        STDOUT: To be used when redirecting stdout.
        STDERR: To be used when redirecting stderr.
    """

    STDOUT = auto()
    STDERR = auto()


class PipeWriter(Protocol):
    """Protocol for a pipe writer"""

    def write(self, data: bytes):
        ...

    def fileno(self) -> int:
        ...

    def close(self):
        ...

    def flush(self):
        ...


# Abstract Pipe class
class ReadWritePipe(ABC):
    """
    Abstract base class for a read-write pipe.

    This class defines the interface for a pipe that can be used for both reading and writing data.
    Subclasses must implement the abstract methods defined here to provide the necessary functionality
    to be used as a pipe for the LogPipe class.
    """

    @abstractmethod
    def read_is_empty(self) -> bool:
        """
        Check if the read buffer is empty.

        Returns:
            bool: True if the read buffer is empty, False otherwise.
        """
        pass

    @abstractmethod
    def close_read(self):
        """
        Close the read end of the pipe.
        """
        pass

    @abstractmethod
    def read_is_closed(self) -> bool:
        """
        Check if the read end of the pipe is closed.

        Returns:
            bool: True if the read end of the pipe is closed, False otherwise.
        """
        pass

    @abstractmethod
    def read(self) -> bytes:
        """
        Read data from the pipe.

        Returns:
            bytes: The data read from the pipe.
        """
        pass

    @abstractmethod
    def close_write(self):
        """
        Close the write end of the pipe.
        """
        pass

    @abstractmethod
    def write_is_closed(self) -> bool:
        """
        Check if the write end of the pipe is closed.

        Returns:
            bool: True if the write end of the pipe is closed, False otherwise.
        """
        pass

    @abstractmethod
    def get_writer(self) -> PipeWriter:
        """
        Get a PipeWriter object for writing to the pipe.

        Returns:
            PipeWriter: A PipeWriter object for writing to the pipe.
        """
        pass


class FileDescriptorReadWritePipe(ReadWritePipe):
    """
    A class that represents a pipe for reading and writing data using standard OS file descriptors.
    These pipes can be used to communicate between processes and are created using the os.pipe() function.

    The class is created to fulfill the need for logging from a python subprocess created with the `subprocess` module.

    Attributes:
        fd_read (int): The file descriptor for reading data.
        fd_write (int): The file descriptor for writing data.
        reader (file object): The file object for reading data.
        _writer_is_closed (bool): A flag indicating whether the writer has been closed.
    """

    def __init__(self):
        super().__init__()
        self.fd_read, self.fd_write = os.pipe()
        self.reader = os.fdopen(self.fd_read, mode="rb")
        self.writer = os.fdopen(self.fd_write, mode="wb")
        self._writer_is_closed = False

    def read_is_empty(self) -> bool:
        return not self.reader.peek()

    def read_is_closed(self) -> bool:
        return self.reader.closed

    def write_is_closed(self) -> bool:
        return self._writer_is_closed

    def close_write(self):
        os.close(self.fd_write)
        self._writer_is_closed = True

    def close_read(self):
        self.reader.close()

    def get_writer(self) -> PipeWriter:
        if self.write_is_closed():
            raise ValueError(
                "Attempted to get writer while the write end of the pipe is closed."
            )
        return self.writer

    def read(self) -> bytes:
        return self.reader.readline()


class MPSpawnReadWritePipe:
    """
    A multiprocessing pipe that allows for reading and writing data between processes.

    This class is a subclass of `ReadWritePipe` and uses the `multiprocessing` module to create a
    pipe that can be used to communicate between processes. It provides methods for reading and
    writing data to the pipe, as well as checking if the pipe is closed or empty. The class is created
    to fulfill the need for logging from a python subprocess created with the `multiprocessing` module using the `spawn` start method.

    Attributes:
        read_conn (multiprocessing.connection.Connection): The connection used for reading data
            from the pipe.
        write_conn (multiprocessing.connection.Connection): The connection used for writing data
            to the pipe.
        _writer_is_closed (bool): A flag indicating whether the writer connection has been closed.
    """

    @dataclass
    class _MPWriter:
        """
        A helper class used to write data to the pipe.

        This class is used internally by `MPSpawnReadWritePipe` to write data to the pipe. It
        implements the `write` and `fileno` methods required by `PipeWriter`.

        Attributes:
            write_conn (multiprocessing.connection.Connection): The connection used for writing
                data to the pipe.
        """

        write_conn: multiprocessing.connection.Connection

        def write(self, data: bytes):
            self.write_conn.send_bytes(data.encode("utf-8"))

        def fileno(self) -> int:
            return self.write_conn.fileno()

        def close(self):
            self.write_conn.close()

        def flush(self):
            pass

    def __init__(self) -> None:
        super().__init__()
        self.read_conn, self.write_conn = multiprocessing.get_context("spawn").Pipe(
            duplex=True
        )
        self._writer = MPSpawnReadWritePipe._MPWriter(write_conn=self.write_conn)
        self._writer_is_closed = False

    def read_is_empty(self) -> bool:
        return not self.read_conn.poll()

    def read_is_closed(self) -> bool:
        return self.read_conn.closed

    def close_read(self):
        return self.read_conn.close()

    def read(self) -> bytes:
        try:
            return self.read_conn.recv_bytes()
        except EOFError:
            return b""

    def write_is_closed(self) -> bool:
        return self._writer_is_closed

    def close_write(self):
        self.write_conn.close()
        self._writer_is_closed = True

    def get_writer(self) -> PipeWriter:
        if self.write_is_closed():
            raise ValueError(
                "Attempted to get writer while the write end of the pipe is closed."
            )
        return self._writer


@contextmanager
def _redirect_stream(stream: StreamEnum, pipe_writer: MPSpawnReadWritePipe._MPWriter):
    """Context manager to redirect stdout or stderr to a pipe writer."""
    # Import sys here to ensure we use the child process sys module

    import sys

    try:
        if stream == StreamEnum.STDOUT:
            sys.stdout = pipe_writer
        elif stream == StreamEnum.STDERR:
            sys.stderr = pipe_writer
        yield
    finally:
        if stream == StreamEnum.STDOUT:
            sys.stdout = sys.__stdout__
        elif stream == StreamEnum.STDERR:
            sys.stderr = sys.__stderr__


def _wrap_to_redirect_stream(func, stream, pipe_writer, *args, **kwargs):
    """Wrap a function to redirect its stdout or stderr to a pipe writer.
    This should be used with functools.partial to keep things pickleable, rather
    than defining the wrapper locally within wrap_and_redirect_stream."""
    with _redirect_stream(stream, pipe_writer):
        return func(*args, **kwargs)


class LogPipe:
    """
    A class to log data from python subprocesses.

    This class is a context manager that can be used to redirect logs from a python subprocess to the
    provided logger. It uses internally a pipe to read data from the either one of the subprocesses's
    standard streams (stdout, stderr) and log it to the provided logger at the provided level.
    The class is created to fulfill the need for logging from a python subprocess.


    Args:
        logger (logging.Logger): The logger to use for logging.
        log_level (int): The log level to use for logging.
        mode (LogPipeMode): The mode to use for the pipe.
        buffer_logs (bool): Whether to buffer the logs in memory. Defaults to False.
            If set the true, the logs will be available in the `log_buffer` attribute after the context manager exits.
            Otherwise, the `log_buffer` attribute will be None.

    Raises:
        ValueError: If an invalid mode is provided.

    Example for the `subprocessing` module:
        ```
        logger = logging.getLogger("foo")
        with LogPipe(logger=logger, log_level=logging.INFO, mode=LogPipeMode.MP_SPAWN, buffer_logs=True) as lp:
            # Run a subprocess and log its output. The stdout of the subprocess will be logged to the logger
            # at the provided level automatically.
            subprocess.run(["echo", "Hello, world!"], stdout=lp.get_pipe_writer())
        # Get the logs from the buffer. The logs will be available in the buffer only if `buffer_logs` was set to True.
        # This is a convenience method only, and could have been done manually by adding a handler to the logger.
        captured_stdout = lp.log_buffer.getvalue()
        ```

    Example for the `multiprocessing` module with the `spawn` start method:
    ```

        def foo(x: int, y: int) -> int:
            print("HELLO FROM FOO")
            return x + y

        logger = logging.getLogger("foo")
        with LogPipe(logger=logger, log_level=logging.INFO, mode=LogPipeMode.MP_SPAWN, buffer_logs=True) as lp:
            # Create a process and log its output. The stdout of the process will be logged to the logger
            # at the provided level automatically.
            foo_wrapped = LogPipe.wrap_and_redirect_stream(foo, StreamEnum.STDOUT, lp.get_pipe_writer())
            p = multiprocessing.get_context("spawn").Process(target=foo_wrapped, args=(lp,))
            p.start()
            p.join()
        # Get the logs from the buffer. The logs will be available in the buffer only if `buffer_logs` was set to True.
        # This is a convenience method only, and could have been done manually by adding a handler to the logger.
        captured_stdout = lp.log_buffer.getvalue()
    """

    def __init__(
        self,
        logger: logging.Logger,
        log_level: int,
        mode: LogPipeMode,
        buffer_logs: bool = False,
    ):
        if mode == LogPipeMode.MP_SPAWN:
            self.pipe = MPSpawnReadWritePipe()
        elif mode == LogPipeMode.SUBPROCESS:
            self.pipe = FileDescriptorReadWritePipe()
        else:
            raise ValueError(
                f"Invalid mode {mode}, expected one of {LogPipeMode.__members__}"
            )
        self.mode = mode
        self.thread = threading.Thread(target=self.run, daemon=False)
        self.logger = logger
        self.log_level = log_level
        self.pipe_read_lock = threading.Lock()
        self.log_buffer = None
        if buffer_logs:
            self.log_buffer = io.StringIO()
            logger.addHandler(logging.StreamHandler(self.log_buffer))

    def log_from_pipe(self, data: bytes):
        data_str = data.decode("utf-8").strip("\n")
        if data_str != "":
            self.logger.log(self.log_level, data_str)

    def run(self):
        """Log everything that comes from the pipe."""
        while True:
            # Check if the write end is closed
            if self.pipe.write_is_closed():
                # Read until empty, then release then lock
                while data := self.pipe.read():
                    self.log_from_pipe(data)
                return
            # Check if not empty and read
            if not self.pipe.read_is_empty():
                # Acquire the lock to ensure the write end isn't closed at this time
                with self.pipe_read_lock:
                    data = self.pipe.read()
                self.log_from_pipe(data)

    def __enter__(self):
        self.thread.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # Close the write end of the pipe
        with self.pipe_read_lock:
            self.pipe.close_write()
        # Wait for the thread to finish and flush any remaining logs
        # TODO(roulbac): Daemon thread, do we really need to join?
        self.thread.join()
        # Close the read end
        self.pipe.close_read()

    def get_pipe_writer(self) -> PipeWriter:
        return self.pipe.get_writer()

    @staticmethod
    def wrap_and_redirect_stream(
        func: Callable, stream: StreamEnum, pipe_writer: PipeWriter
    ) -> Callable:
        """
        Wraps a function to redirect its output to a pipe writer.

        Args:
            func (Callable): The function to wrap.
            stream (Literal["stdout", "stderr"]): The stream to redirect. Must be one of "stdout" or "stderr".
            pipe_writer (PipeWriter): The pipe writer to redirect the stream to.

        Returns:
            Callable: The wrapped function.
        """

        if stream not in StreamEnum.__members__.values():
            raise ValueError(
                f"Invalid stream {stream}, expected one of {StreamEnum.__members__}"
            )

        if not isinstance(pipe_writer, MPSpawnReadWritePipe._MPWriter):
            raise ValueError(
                "Invalid pipe_writer, please provide a writer for a LogPipe started in MP_SPAWN mode"
            )

        return partial(
            _wrap_to_redirect_stream, func=func, stream=stream, pipe_writer=pipe_writer
        )



from typing import Any, Union


class NodeReference:
    """Node parameter type"""

    ref: str

    def __init__(self, ref: str) -> None:
        self.ref = ref

    def unwrap(self):
        return self.ref


class Array:
    """Node parameter type of `list[str | float | int]`"""

    ref: list[str | float | int]

    def __init__(self, ref: list[str | float | int]) -> None:
        self.ref = ref

    def unwrap(self):
        return self.ref


def format_param_value(value: Any, value_type: str):
    match value_type:
        case "Array":
            s = str(value)
            parsed_value = parse_array(s, [str, float, int], "list[int | float | str]")
            return Array(parsed_value)
        case "float":
            return float(value)
        case "int":
            return int(value)
        case "bool":
            return bool(value)
        case "NodeReference":
            return NodeReference(str(value))
        case "list[str]":
            return parse_array(str(value), [str], "list[str]")
        case "list[float]":
            return parse_array(str(value), [float], "list[float]")
        case "list[int]":
            return parse_array(str(value), [int], "list[int]")
        case "select" | "str":
            return str(value)
        case _:
            return value


def parse_array(
    str_value: str, type_list: list[Any], param_type: str
) -> list[Union[int, float, str]]:
    if not str_value:
        return []

    val_list = [val.strip() for val in str_value.split(",")]
    # First try to cast into int, then float, then keep as string if all else fails
    for t in type_list:
        try:
            return list(map(t, val_list))
        except ValueError:
            continue

    raise ValueError(
        f"Couldn't parse list items with type {','.join([str(t) for t in type_list])}."
        + f"Value should be comma (',') separated {' | '.join([t.__name__ for t in type_list])} for parameter type {param_type}."
    )
